/*
 * Copyright (c) 2010 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include "velodyne3DImage.h"

/* System headers */
#include <math.h>
#include <stdio.h>
#include <string.h>

/* External headers */
#include <t3d/t3d.h>

/* Genom headers */
#include <portLib.h>

/* Internal headers */
#include "server/velodyneHeader.h"
#include "velodyneBuffer.h"

/*
 * Completes the columns organization. The image resulting from the processing
 * of all the points of a shot contains empty elements in its leftmost region
 * due to the column offsets. It also contains points in columns to the right of
 * the region determined by the scan width, again due to the column offsets.
 * However, the leftmost points and the rightmost points are actually neighbors,
 * and thus the rightmost points should be moved to the left. This function
 * performs this operation.
 */
void
velodyneCompleteColumnsOrganization(velodyne3DImage* image,
                                    velodyne3DImageParams* imageParams,
                                    int scanWidth)
{
  int i;
  velodyne3DPoint * ptDest, * ptSource;
  size_t size;
  for (i = 0; i < image->height; i++) {
    if (imageParams->columnOffsets[i] >= 1) {
      ptDest = velodyne3DImageAt(image, i, 0);
      ptSource = velodyne3DImageAt(image, i, scanWidth);
      size = imageParams->columnOffsets[i] * sizeof (velodyne3DPoint);
      memcpy(ptDest, ptSource, size);
      memset(ptSource, 0, size);
    }
  }
}

/*
 * Fill the line of the image with the points of a cartesian packet
 */
static inline STATUS
velodyneAddPacket(velodyne3DImage* image,
                  velodyne3DImageParams* imageParams,
                  VELODYNE_CARTESIAN_PACKET* packet,
                  POM_SENSOR_POS* currentPos,
                  POM_SENSOR_POS* startPos,
                  int numPacketsInScan,
                  velodyneTargetDetectionParams* detectionParams,
                  velodyneTargetParams* targetParams,
                  velodyneTarget* target)
{
  int i, j;
  T3D sensor2ToMain2;
  T3D main2ToOrigin;
  T3D sensor2ToOrigin;
  T3D main1ToOrigin;
  T3D originToMain1;
  T3D sensor2ToMain1;
  T3D main1ToSensor1;
  T3D sensor2ToSensor1;
  /* The suffix 2 stands for 'current', 1 stands for 'start' */

  int line, col;
  velodyne3DPoint* imgPoint;
  int packetIndex = 0;
  short intensity;
  double x, y, z;
  float xF, yF, zF;
  velodyne3DPoint* targetPoint;
  double* c;
  t3dInit (&sensor2ToMain2, T3D_BRYAN, T3D_ALLOW_CONVERSION);
  t3dInit (&main2ToOrigin, T3D_BRYAN, T3D_ALLOW_CONVERSION);
  t3dInit (&sensor2ToOrigin, T3D_BRYAN, T3D_ALLOW_CONVERSION);
  t3dInit (&main1ToOrigin, T3D_BRYAN, T3D_ALLOW_CONVERSION);
  t3dInit (&originToMain1, T3D_BRYAN, T3D_ALLOW_CONVERSION);
  t3dInit (&sensor2ToMain1, T3D_BRYAN, T3D_ALLOW_CONVERSION);
  t3dInit (&main1ToSensor1, T3D_BRYAN, T3D_ALLOW_CONVERSION);
  t3dInit (&sensor2ToSensor1, T3D_BRYAN, T3D_ALLOW_CONVERSION);

  sensor2ToMain2.euler.euler[0] = currentPos->sensorToMain.euler.yaw;
  sensor2ToMain2.euler.euler[1] = currentPos->sensorToMain.euler.pitch;
  sensor2ToMain2.euler.euler[2] = currentPos->sensorToMain.euler.roll;
  sensor2ToMain2.euler.euler[3] = currentPos->sensorToMain.euler.x;
  sensor2ToMain2.euler.euler[4] = currentPos->sensorToMain.euler.y;
  sensor2ToMain2.euler.euler[5] = currentPos->sensorToMain.euler.z;

  main2ToOrigin.euler.euler[0] = currentPos->mainToOrigin.euler.yaw;
  main2ToOrigin.euler.euler[1] = currentPos->mainToOrigin.euler.pitch;
  main2ToOrigin.euler.euler[2] = currentPos->mainToOrigin.euler.roll;
  main2ToOrigin.euler.euler[3] = currentPos->mainToOrigin.euler.x;
  main2ToOrigin.euler.euler[4] = currentPos->mainToOrigin.euler.y;
  main2ToOrigin.euler.euler[5] = currentPos->mainToOrigin.euler.z;

  t3dCompIn (&sensor2ToOrigin, &sensor2ToMain2, &main2ToOrigin);

  main1ToOrigin.euler.euler[0] = startPos->mainToOrigin.euler.yaw;
  main1ToOrigin.euler.euler[1] = startPos->mainToOrigin.euler.pitch;
  main1ToOrigin.euler.euler[2] = startPos->mainToOrigin.euler.roll;
  main1ToOrigin.euler.euler[3] = startPos->mainToOrigin.euler.x;
  main1ToOrigin.euler.euler[4] = startPos->mainToOrigin.euler.y;
  main1ToOrigin.euler.euler[5] = startPos->mainToOrigin.euler.z;

  t3dInvertIn (&originToMain1, &main1ToOrigin);
  t3dCompIn (&sensor2ToMain1, &sensor2ToOrigin, &originToMain1);

  t3dInvertIn (&main1ToSensor1, &sensor2ToMain2);
      /* Note: sensor2ToMain2 = sensor1ToMain1 */
  t3dCompIn (&sensor2ToSensor1, &sensor2ToMain1, &main1ToSensor1);

  if (!t3dOk (&sensor2ToSensor1)) {
    fprintf (stderr, "velodyne: fillLineWithPacket: bad t3d\n");
    return ERROR;
  }

  if (!t3dConvertTo (T3D_MATRIX, &sensor2ToSensor1)) {
    fprintf (stderr, "velodyne: fillLineWithPacket: can't convert t3d\n");
    return ERROR;
  }

  for (j = 0; j < imageParams->numColumnsPerPacket; j++) {
    for (i = 0; i < imageParams->numLinesPerPacket; i++) {
      if (imageParams->organizationEnabled == VELODYNE_TRUE) {
        line = imageParams->lineOrder[i];
        col = j + numPacketsInScan * imageParams->numColumnsPerPacket
              + imageParams->columnOffsets[line];
      } else {
        line = i;
        col = j + numPacketsInScan * imageParams->numColumnsPerPacket;
      }
      if (packet->points[packetIndex].status == VELODYNE_GOOD_POINT) {
        c = sensor2ToSensor1.matrix.matrix;
        x = (double) packet->points[packetIndex].x;
        y = (double) packet->points[packetIndex].y;
        z = (double) packet->points[packetIndex].z;
        intensity = (short) packet->points[packetIndex].intensity;
        imgPoint = velodyne3DImageAt(image, line, col);
        imgPoint->coordinates[0]
            = (float) (x * *c + y * *(c + 1) + z * *(c + 2) + *(c + 3));
        c += 4;
        imgPoint->coordinates[1]
            = (float) (x * *c + y * *(c + 1) + z * *(c + 2) + *(c + 3));
        c += 4;
        imgPoint->coordinates[2]
            = (float) (x * *c + y * *(c + 1) + z * *(c + 2) + *(c + 3));
        imgPoint->intensity = intensity;
        imgPoint->status = VELODYNE_GOOD_3DPOINT;

        if (detectionParams->detectionOn == VELODYNE_TRUE
            && intensity >= targetParams->minReflectance
            && intensity <= targetParams->maxReflectance) {
          /* We have a point that may correspond to a target. Check that we do
             not exceed the max number of target points allocated, that the
             point is within the detection distance range and that it is over
             the minimum considered target elevation */
          if (target->detectedTargetPoints >= VELODYNE_MAX_TARGET_POINTS) {
            fprintf (stderr,
                     "velodyne: warning, number of target points exceeding the max number (%d)\n",
                     VELODYNE_MAX_TARGET_POINTS);
          } else {
            xF = imgPoint->coordinates[0];
            yF = imgPoint->coordinates[1];
            zF = imgPoint->coordinates[2];
            if (fabs (x) < targetParams->maxDistance
                && fabs (y) < targetParams->maxDistance
                && fabs (z) < targetParams->maxDistance
                && z > targetParams->zMin) {
              targetPoint = target->targetPoints + target->detectedTargetPoints;
              targetPoint->status = VELODYNE_GOOD_3DPOINT;
              targetPoint->intensity = intensity;
              targetPoint->coordinates[0] = xF;
              targetPoint->coordinates[1] = yF;
              targetPoint->coordinates[2] = zF;
              target->detectedTargetPoints += 1;
            }
          }
        }
      }
      packetIndex++;
    }
  }
  return OK;
}

/*
 * Extracts one scan from a velodyne stream. Each packet is a set
 * VELODYNE_NUM_SHOT (=12) shots, with angles attached. The angles are encoded
 * in 100th of degrees (range from 0 to 35999).
 *
 * The angle of the first shot from each packet is used to check that a full
 * scan has been acquired (one could have checked the angles of each shot
 * within a packet, but this becomes a bit tedious -- and not really necessary
 * after all).
 */
ACTIVITY_EVENT
velodyneGrabOneShot(velodyneBuffer* buffer,
                    velodyneOneShotParams* oneShotParams,
                    velodyne3DImage* image,
                    velodyne3DImageParams* imageParams,
                    velodyneTargetDetectionParams* detectionParams,
                    velodyneTargetParams* targetParams,
                    velodyneTarget* target)
{
  VELODYNE_PACKET* packet;
  VELODYNE_CARTESIAN_PACKET cartPacket;
  int currentHeading;
  int previousHeading; /* To check the progression of the heading angle */
  int numPacketsInScan = 0;
  int deltaHeading = 0; /* The angle spanned since the first packet read
                           (encoded in 100th of degrees) */
  POM_SENSOR_POS* startPos = NULL;
  POM_SENSOR_POS* currentPos = NULL;
  int scanWidth;

  /* The following parameters are used to specify which packets must be put in
     the scan */
  double yawMainToSensor; /* The heading angle between sensor and robot */
  int firstAngleVelodyne;
  int lastAngleVelodyne;
  int reversedAngles; /* Flag to indicate the first angle is the smallest or
                         not. */
  /* int numAddedPackets = 0; /* The number of packets in the specified angular */
  /*                             sector * / */

  if (!buffer->dataOverwriting && buffer->read == buffer->write)
    return END;
  if (buffer->dataOverwriting)
    buffer->read = buffer->write;

  /* Read the first packet, its pos and its heading */
  packet = &(buffer->packets[buffer->read].packet);
  currentPos = &(buffer->packets[buffer->read].position);
  startPos = &(buffer->packets[buffer->read].position);
  currentHeading = packet->data.firings[0].rotational_pos;

  /* Computes the heading offset */
  yawMainToSensor = -currentPos->sensorToMain.euler.yaw * 180.0 / M_PI;

  /* Computes angles in the original Velodyne reference frame */
  velodyne_convert_heading_sector_to_original_reference_frame(
      imageParams->model,
      oneShotParams->firstAngle + yawMainToSensor,
      oneShotParams->lastAngle + yawMainToSensor,
      &firstAngleVelodyne,
      &lastAngleVelodyne,
      &reversedAngles);

  /* Add the packet to the image */
  if (reversedAngles != 0) {
    if (currentHeading >= firstAngleVelodyne
        || currentHeading < lastAngleVelodyne) {
      velodyne_to_cartesian_packet (imageParams->model, packet, &cartPacket);
      if (velodyneAddPacket (
              image, imageParams, &cartPacket, currentPos, startPos,
              numPacketsInScan, detectionParams, targetParams, target) != OK)
        return ETHER;
      /* numAddedPackets++; */
    }
  } else {
    if (currentHeading >= firstAngleVelodyne
        && currentHeading < lastAngleVelodyne) {
      velodyne_to_cartesian_packet (imageParams->model, packet, &cartPacket);
      if (velodyneAddPacket (
              image, imageParams, &cartPacket, currentPos, startPos,
              numPacketsInScan, detectionParams, targetParams, target) != OK)
        return ETHER;
      /* numAddedPackets++; */
    }
  }
  buffer->read = (buffer->read + 1) % buffer->nbPackets;
  if (buffer->dataOverwriting)
    buffer->dataOverwriting = false;
  numPacketsInScan++;

  /* Now loop until a full 360 degrees scan has been read */
  while (deltaHeading < 36000) {
    if (buffer->read == buffer->write) /* Read the packet only if it is not
                                          being updated by the acquisition
                                          task */
      goto leavingOut;

    if (numPacketsInScan >= buffer->nbPackets) {
      fprintf (stderr,
               "velodyne: warning, number of packets exceeding poster size\n");
      return ETHER;
    }

    /* Get the current packet, its pos and heading */
    packet = &(buffer->packets[buffer->read].packet);
    currentPos = &(buffer->packets[buffer->read].position);
    previousHeading = currentHeading;
    currentHeading = packet->data.firings[11].rotational_pos;

    /* Update deltaHeading */
    if (currentHeading < previousHeading)
      /* In this case we crossed the 0 angle between previousHeading and
         currentHeading */
      deltaHeading += currentHeading + 36000 - previousHeading;
    else
      deltaHeading += currentHeading - previousHeading;

    /* Add the packet to the image */
    if (reversedAngles != 0) {
      if (currentHeading > firstAngleVelodyne
          || currentHeading <= lastAngleVelodyne) {
        velodyne_to_cartesian_packet (imageParams->model, packet, &cartPacket);
        if (velodyneAddPacket (
                image, imageParams, &cartPacket, currentPos, startPos,
                numPacketsInScan, detectionParams, targetParams, target) != OK)
          return ETHER;
        /* numAddedPackets++; */
      }
    } else {
      if (currentHeading > firstAngleVelodyne
          && currentHeading <= lastAngleVelodyne) {
        velodyne_to_cartesian_packet (imageParams->model, packet, &cartPacket);
        if (velodyneAddPacket (
                image, imageParams, &cartPacket, currentPos, startPos,
                numPacketsInScan, detectionParams, targetParams, target) != OK)
          return ETHER;
        /* numAddedPackets++; */
      }
    }
    buffer->read = (buffer->read + 1) % buffer->nbPackets;
    numPacketsInScan++;
  }

  leavingOut:
    /* Sets the number of used columns */
    scanWidth = numPacketsInScan * imageParams->numColumnsPerPacket;
    if (imageParams->organizationEnabled == VELODYNE_TRUE
        && imageParams->model == VELODYNE_64)
      velodyneCompleteColumnsOrganization (image, imageParams, scanWidth);
    image->width = scanWidth;

    /* Writes the position in the image */
    memcpy (&(image->position), startPos, sizeof (POM_SENSOR_POS));

    /* printf ("Nb packets in the specified sector: %d (out of %d total packets)\n", */
    /* numAddedPackets, numPacketsInScan); */

  return END;
}
