/*
 * Copyright (c) 2010 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include "velodyneBuffer.h"

/* System headers */
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

/* Public variables */
extern VELODYNE_INTERFACE* interface;
extern velodyneReadTask task;
extern POSTER_ID sensorPosPosterId;

void
velodyneCreateBuffer(int numPackets)
{
  task.buffer.nbPackets = numPackets;
  task.buffer.read = 0;
  task.buffer.write = 0;
  task.buffer.dataOverwriting = false;
  task.buffer.packets = malloc (numPackets * sizeof (velodynePacket));
  if (task.buffer.packets == NULL)
    fprintf (stderr, "velodyne: cannot allocate buffer for %d packets\n",
             numPackets);
}

void
velodyneDestroyBuffer()
{
  if (task.buffer.packets != NULL) {
    free (task.buffer.packets);
    task.buffer.packets = NULL;
  }
  task.buffer.nbPackets = 0;
  task.buffer.read = 0;
  task.buffer.write = 0;
  task.buffer.dataOverwriting = false;
}

/*
 * XXX Unlocked access to SDI_F->running, but we can guess it is atomic
 * XXX No way to return a useful error to system as the thread run in
 * background
 */
void*
velodyneReadLive()
{
  static POM_SENSOR_POS pos;
  static int i = 0;

  while (SDI_F->running != 0) {
    pthread_mutex_lock (&task.mtx);
    if (task.needOneShot)
      pthread_cond_wait (&task.cond, &task.mtx);

    if (velodyne_read_packet (interface,
                              &(task.buffer.packets[task.buffer.write].packet))
        != VELODYNE_PAYLOAD_SIZE) {
      fprintf (stderr, "velodyne: cannot read packet\n");
      SDI_F->running = 0;
      pthread_mutex_unlock (&task.mtx);
      return NULL;
    }
    if (SDI_F->initParams.usePom == VELODYNE_TRUE) {
      if (i++ % 50 == 0) {
        if (velodynePOM_SENSOR_POSPosterRead (sensorPosPosterId, &pos)
            == ERROR) {
          fprintf (stderr, "velodyne: unable to read pom poster\n");
          SDI_F->running = 0;
          pthread_mutex_unlock (&task.mtx);
          return NULL;
        }
      }
      task.buffer.packets[task.buffer.write].position = pos;
    }

    task.buffer.write = (task.buffer.write + 1) % task.buffer.nbPackets;
    if (task.buffer.write == task.buffer.read)
      task.buffer.dataOverwriting = true;
    pthread_mutex_unlock (&task.mtx);
  }

  return NULL;
}

int
velodynePosFileOpen(const char* fileName, FILE** fileId)
{
  char aux[VELODYNE_MAX_STRLEN];
  FILE* f = fopen(fileName, "r");
  if (f == NULL) {
    *fileId = NULL;
    return VELODYNE_READER_ERR_CANNOT_OPEN_POS_FILE;
  }
  if (fscanf(f, "%s\t%s\t%s\t%s\t%s\t%s\t%s\n",
             aux, aux, aux, aux, aux, aux, aux)
      != 7) {
    fclose(f);
    *fileId = NULL;
    return VELODYNE_READER_ERR_BAD_POS_FILE_FORMAT;
  }
  *fileId = f;
  return VELODYNE_READER_OK;
}

int
velodynePosFileRead(FILE* fileId, double* timespec, POM_EULER* pos)
{
  int result;
  result = fscanf(fileId, "%lf\t%f\t%f\t%f\t%f\t%f\t%f\n",
                  timespec,
                  &(pos->yaw), &(pos->pitch), &(pos->roll),
                  &(pos->x), &(pos->y), &(pos->z));
  if (result == EOF)
    return VELODYNE_READER_ERR_END_POS_FILE;
  else if (result != 7)
    return VELODYNE_READER_ERR_CANNOT_READ_POS_FILE;
  return VELODYNE_READER_OK;
}

void*
velodyneReadPlayback(void* positionFileId)
{
  FILE* posFileId = (FILE*) positionFileId;
  double tCurr, tNext, tPacket;
  POM_EULER mainToOriginCurr, mainToOriginNext;
  bool oneShotRequired = false;
  double errCurr, errNext;
  int posFileReadResult;

  /* Initializes current position. */
  posFileReadResult = velodynePosFileRead(posFileId, &tCurr, &mainToOriginCurr);
  if (posFileReadResult == VELODYNE_READER_ERR_END_POS_FILE) {
    fprintf(stderr, "velodyne: end of position file\n");
    SDI_F->running = 0;
    return NULL;
  } else if (posFileReadResult == VELODYNE_READER_ERR_CANNOT_READ_POS_FILE) {
    fprintf(stderr, "velodyne: cannot read position file\n");
    SDI_F->running = 0;
    return NULL;
  }

  /* Initializes next position. */
  posFileReadResult = velodynePosFileRead(posFileId, &tNext, &mainToOriginNext);
  if (posFileReadResult == VELODYNE_READER_ERR_END_POS_FILE) {
    fprintf(stderr, "velodyne: end of position file\n");
    SDI_F->running = 0;
    return NULL;
  } else if (posFileReadResult == VELODYNE_READER_ERR_CANNOT_READ_POS_FILE) {
    fprintf(stderr, "velodyne: cannot read position file\n");
    SDI_F->running = 0;
    return NULL;
  }

  /* Reading loop. */
  while (SDI_F->running != 0) {
    /* Pre-reading operations. */
    pthread_mutex_lock(&task.mtx);
    if (task.buffer.dataOverwriting)
      pthread_cond_wait(&task.cond, &task.mtx);

    /* Puts Velodyne packet in the buffer. */
    if (velodyne_read_packet(
            interface, &(task.buffer.packets[task.buffer.write].packet))
        != VELODYNE_PAYLOAD_SIZE) {
      if (velodyne_read_packet(
              interface, &(task.buffer.packets[task.buffer.write].packet))
          != VELODYNE_PAYLOAD_SIZE) {
        fprintf(stderr, "velodyne: cannot read packet\n");
        SDI_F->running = 0;
        pthread_mutex_unlock(&task.mtx);
        return NULL;
      }
    }

    /* Computes initial timespec errors for the packet. */
    tPacket
        = task.buffer.packets[task.buffer.write].packet.ts.tv_sec
          + task.buffer.packets[task.buffer.write].packet.ts.tv_usec * 1.0E-6;
    errCurr = fabs(tPacket - tCurr);
    errNext = fabs(tPacket - tNext);

    /* Searches the position that corresponds to the packet. */
    while (errNext < errCurr) {
      tCurr = tNext;
      mainToOriginCurr = mainToOriginNext;
      errCurr = errNext;
      posFileReadResult = velodynePosFileRead(posFileId, &tNext,
                                              &mainToOriginNext);
      if (posFileReadResult == VELODYNE_READER_ERR_END_POS_FILE) {
        fprintf(stderr, "velodyne: end of position file\n");
        SDI_F->running = 0;
        pthread_mutex_unlock(&task.mtx);
        return NULL;
      } else if (posFileReadResult
                 == VELODYNE_READER_ERR_CANNOT_READ_POS_FILE) {
        fprintf(stderr, "velodyne: cannot read position file\n");
        SDI_F->running = 0;
        pthread_mutex_unlock(&task.mtx);
        return NULL;
      }
      errNext = fabs(tPacket - tNext);
    }

    /* Puts position in the buffer. */
    task.buffer.packets[task.buffer.write].position.sensorToMain.euler
        = SDI_F->sensorToMain;
    task.buffer.packets[task.buffer.write].position.mainToOrigin.euler
        = mainToOriginCurr;

    /* Updates buffer's write pointer. */
    task.buffer.write = (task.buffer.write + 1) % task.buffer.nbPackets;

    /* Post-reading operations. */
    if (task.buffer.write == task.buffer.read)
      task.buffer.dataOverwriting = true;
    pthread_mutex_unlock(&task.mtx);
  }
  return NULL;
}
